import asyncio
from enum import Enum
from pathlib import Path
from typing import Literal

from typing_extensions import LiteralString

from piou import Cli, Option, Password, Secret
from piou.utils import Regex
from piou.exceptions import CommandError
from piou.formatter import RichFormatter

cli = Cli(description="A Cli tool", formatter=RichFormatter(show_default=True))


@cli.processor()
def processor(
    quiet: bool = Option(False, "-q", "--quiet", help="Do not output any message"),
    verbose: bool = Option(False, "--verbose", help="Increase verbosity"),
):
    print(f"Processing {quiet=} and {verbose=}")


LongLiteral = Literal["foo1", "foo2", "foo3", "foo4"]
UnionLiteral = Literal["foo1"] | Literal["foo2"]


class AnEnum(Enum):
    value_1 = "value-1"
    value_2 = "value-2"


@cli.command(cmd="foo", help="Run foo command")
def foo_main(
    foo1: int = Option(..., help="Foo argument"),
    foo2: str = Option(..., "-f", "--foo2", help="Foo2 argument"),
    foo3: str | None = Option(None, "-g", "--foo3", help="Foo3 argument"),
    foo4: Literal["foo", "bar"] | None = Option(None, "--foo4", help="Foo4 argument", case_sensitive=False),
    foo5: list[int] | None = Option(None, "--foo5", help="Foo5 arguments"),
    foo6: LongLiteral = Option("foo1", "--foo6", help="Foo6 argument"),
    foo7: LongLiteral | None = Option(None, "--foo7", help="Foo7 argument"),
    foo8: UnionLiteral | None = Option(None, "--foo8", help="Foo8 argument"),
    foo9: str | None = Option(None, "--foo9", help="Foo9 argument", choices=[f"item{i}" for i in range(4)]),
    foo10: AnEnum = Option(AnEnum.value_1, "--foo10", help="Foo10 argument"),
    foo11: LiteralString | None = Option(None, "--foo11", help="Foo11 argument"),
    foo12: dict | None = Option(None, "--foo12", help="Foo12 argument"),
    foo13: int | None = Option(None, "--foo13", help="Foo13 argument", choices=lambda: list(range(10))),
    foo14: str | None = Option(None, "--foo14", help="Foo14 argument", choices=["prod", "staging", Regex(r"dev-\d+")]),
    foo_bar: int = Option(..., "--foo-bar", help="Foo bar argument"),
):
    """
    Example command
    """
    print("Running foo")
    for name, value in [
        ("foo1", foo1),
        ("foo2", foo2),
        ("foo3", foo3),
        ("foo4", foo4),
        ("foo5", foo5),
        ("foo6", foo6),
        ("foo7", foo7),
        ("foo8", foo8),
        ("foo9", foo9),
        ("foo10", foo10),
        ("foo11", foo11),
        ("foo12", foo12),
        ("foo_bar", foo_bar),
        ("foo13", foo13),
        ("foo14", foo14),
    ]:
        print(f"{name} = {value} ({type(value)})")


@cli.command(cmd="bar", help="Run bar command")
def bar_main(**kwargs):
    pass


@cli.command(cmd="error", help="Raise an error")
def error_main():
    raise CommandError("An error occurred")


sub_cmd = cli.add_sub_parser(cmd="sub", help="A sub command", propagate_options=True)
sub_cmd.add_option("--test", help="Test mode")


@sub_cmd.command(cmd="bar", help="Run bar command")
def sub_bar_main(foo: Path = Option(Path("/tmp"), "--foo", help="Foo argument"), **kwargs):
    print("Running sub-bar command")
    print(f"foo ({type(foo)})")


@sub_cmd.command(cmd="foo", help="Run foo command")
def sub_foo_main(
    test: bool,
    foo1: int = Option(..., help="Foo argument"),
    foo2: str = Option(..., "-f", "--foo2", help="Foo2 argument"),
    foo3: str | None = Option(None, "--foo3", help="Foo3 argument"),
):
    for name, value in [("test", test), ("foo1", foo1), ("foo2", foo2), ("foo3", foo3)]:
        print(f"{name} = {value} ({type(value)})")


async def _task(task_id: int):
    print(f"Task {task_id} starting")
    try:
        await asyncio.sleep(3)
    except asyncio.CancelledError:
        print(f"Task {task_id} cancelled")
    else:
        print(f"Task {task_id} done")


@cli.command("async-main", help="Run async main with TaskGroup")
async def _async_main():
    async with asyncio.TaskGroup() as tg:
        for i in range(3):
            tg.create_task(_task(i))


@cli.command(cmd="secrets", help="Run secrets command")
def secrets_main(
    password: Password = Option("my-password", "--password", help="Password (fully masked)"),
    # Option-style masking configuration
    token: Secret = Option("sk-12345678", "--token", help="API token (shows first 3 chars)", show_first=3),
    card: Secret = Option("4111111111111234", "--card", help="Card number (shows last 4 chars)", show_last=4),
):
    """
    Example using Password and Secret types.
    You can run it with:
    ```bash
     python -m examples secrets --help
     python -m examples secrets --password secret123 --token sk-abcdef --card 1234567890
    ```
    """
    print(f"password={password}, token={token}, card={card}")


if __name__ == "__main__":
    try:
        cli.run()
    except KeyboardInterrupt:
        print("Ctrl+C detected, exiting...")



from typing import Literal

from piou import Annotated, Cli, Derived, Option

cli = Cli(description="A CLI tool using Annotated syntax")


@cli.command(cmd="foo", help="Run foo command")
def foo_main(
    bar: Annotated[int, Option(..., help="Bar positional argument (required)")],
    baz: Annotated[str, Option(..., "-b", "--baz", help="Baz keyword argument (required)")],
    foo: Annotated[str | None, Option(None, "--foo", help="Foo keyword argument")],
):
    """
    A longer description on what the function is doing.
    You can run it with:
    ```bash
     python -m examples.annotated foo 1 -b baz
    ```
    And you are good to go!
    """
    print(f"bar={bar}, baz={baz}, foo={foo}")


def processor(a: int = Option(1, "-a"), b: int = Option(2, "-b")) -> int:
    return a + b


@cli.command(cmd="derived", help="Run derived command with Annotated syntax")
def derived_main(
    value: Annotated[int, Derived(processor)],
    mode: Annotated[Literal["debug", "release"], Option("debug", "--mode", help="Run mode")],
):
    """
    Example using Annotated with Derived options.
    You can run it with:
    ```bash
     python -m examples.annotated derived -a 3 -b 2 --mode release
    ```
    """
    print(f"value={value}, mode={mode}")


if __name__ == "__main__":
    cli.run()



import sys
from pathlib import Path
from unittest.mock import patch
from rich.console import Console
import importlib

# Standardize console for all screenshots
CONSOLE_WIDTH = 100
EXPORT_DIR = Path("docs")


def generate_screenshot(module_name, args, output_filename, title="Terminal"):
    print(f"Generating {output_filename}...")

    # Create a recording console
    console = Console(record=True, width=CONSOLE_WIDTH, color_system="truecolor")

    # Mock sys.argv
    # We set argv[0] to module_name + ".py" so it looks like a real script run
    mock_argv = [f"{module_name.split('.')[-1]}.py"] + args

    with patch("sys.argv", mock_argv), patch("piou.formatter.rich_formatter.Console", return_value=console):
        # Reset modules to force a re-import if necessary, or just import once
        if module_name in sys.modules:
            importlib.reload(sys.modules[module_name])
        else:
            importlib.import_module(module_name)

        # Get the 'cli' object from the module
        cli = getattr(sys.modules[module_name], "cli")

        # Run and catch SystemExit (which piou calls on error or help)
        try:
            cli.run()
        except SystemExit:
            pass

    # Save as SVG
    output_path = EXPORT_DIR / output_filename
    console.save_svg(str(output_path), title=title)
    print(f"Saved to {output_path}")


if __name__ == "__main__":
    # Ensure docs directory exists
    EXPORT_DIR.mkdir(exist_ok=True)

    # 1. simple-output.svg
    generate_screenshot("examples.simple", ["-h"], "simple-output.svg", "simple.py -h")

    # 2. simple-output-foo.svg
    generate_screenshot("examples.simple", ["foo", "-h"], "simple-output-foo.svg", "simple.py foo -h")

    # 3. sub-cmd-output.svg
    # The README sub-command example matches examples.__main__ sub command
    generate_screenshot("examples.__main__", ["sub", "-h"], "sub-cmd-output.svg", "run.py sub -h")

    # 4. dynamic-derived.svg
    generate_screenshot("examples.derived", ["dynamic", "-h"], "dynamic-derived.svg", "derived.py dynamic -h")

    print("\nAll screenshots generated as SVG!")



from __future__ import annotations

import os
import re
import sys
from pathlib import Path
from typing import Literal, TYPE_CHECKING, cast

from ..utils import _SecretBase
from ..command import CommandOption

if TYPE_CHECKING:
    from .base import Formatter

FormatterType = Literal["raw", "rich"]


def mask_secret(value: str, show_first: int = 0, show_last: int = 0, replacement: str = "*") -> str:
    """Mask a secret value by replacing hidden characters with `replacement`.

    Optionally keeps `show_first` characters visible at the beginning
    and `show_last` characters at the end.
    """
    if not value:
        return replacement * 6
    total_len = len(value)
    if show_first + show_last >= total_len:
        return value
    prefix = value[:show_first] if show_first > 0 else ""
    suffix = value[-show_last:] if show_last > 0 else ""
    hidden_len = total_len - show_first - show_last
    mask = replacement * min(hidden_len, 6)
    return f"{prefix}{mask}{suffix}"


def get_formatter(formatter_type: FormatterType | None = None) -> Formatter:
    """Return a Formatter instance based on `formatter_type`.

    If None, uses PIOU_FORMATTER env var or defaults to 'rich' if available.
    """
    from .base import Formatter

    if formatter_type is None:
        formatter_type = os.environ.get("PIOU_FORMATTER", "").lower() or None  # type: ignore[assignment]

    if formatter_type == "raw":
        return Formatter()
    elif formatter_type == "rich":
        try:
            from .rich_formatter import RichFormatter

            return RichFormatter()
        except ImportError:
            raise ImportError("Rich is not installed. Install it with: pip install rich")

    # Default: use Rich if available
    try:
        from .rich_formatter import RichFormatter

        return RichFormatter()
    except ImportError:
        return Formatter()


def get_program_name() -> str:
    """
    Get the program name for display in usage messages.

    Handles both direct script execution and module execution (python -m module_name).
    When run as a module, returns the module name instead of __main__.py.
    """
    program_name = os.path.basename(sys.argv[0])
    if program_name == "__main__.py":
        import __main__

        if hasattr(__main__, "__file__") and __main__.__file__:
            module_path = Path(__main__.__file__).parent
            program_name = module_path.name
        else:
            main_module = sys.modules.get("__main__")
            if main_module and hasattr(main_module, "__package__") and main_module.__package__:
                program_name = main_module.__package__

    return program_name


def fmt_option_raw(option: CommandOption, show_full: bool = False) -> str:
    """Format a command option for display (plain text)."""
    if option.is_positional_arg:
        return f"<{option.name}>"
    elif show_full:
        first_arg, *other_args = option.keyword_args
        required = "*" if option.is_required else ""
        if other_args:
            other_args_str = ", ".join(other_args)
            return f"{first_arg} ({other_args_str}){required}"
        else:
            return f"{first_arg}{required}"
    else:
        return "[" + sorted(option.keyword_args)[-1] + "]"


def fmt_cmd_options_raw(options: list[CommandOption]) -> str:
    """Format command options for display in the usage section (plain text)."""
    return " ".join([fmt_option_raw(x) for x in options]) if options else ""


def fmt_choice(choice: str | re.Pattern[str]) -> str:
    """Format a choice for display, converting regex patterns to /pattern/ format."""
    if isinstance(choice, re.Pattern):
        return f"/{choice.pattern}/"
    return str(choice)


def fmt_help(
    option: CommandOption,
    show_default: bool,
    *,
    markdown_open: str | None = None,
    markdown_close: str | None = None,
) -> str | None:
    """Format the help text for a command option."""
    _choices = option.get_choices()
    _markdown_open, _markdown_close = markdown_open or "", markdown_close or ""

    if show_default and option.default is not None and not option.is_required:
        if option.is_secret:
            if option.show_first is not None or option.show_last is not None:
                _show_first, _show_last, _replacement = (
                    option.show_first or 0,
                    option.show_last or 0,
                    option.replacement,
                )
            else:
                _secret = cast(type[_SecretBase], option.data_type)
                _show_first, _show_last, _replacement = _secret.show_first, _secret.show_last, _secret.replacement
            default_str = mask_secret(str(option.default), _show_first, _show_last, _replacement)
        else:
            default_str = option.default
        default_str = f"{_markdown_open}(default: {default_str}){_markdown_close}"
        return option.help + f" {default_str}" if option.help else default_str
    elif _choices is not None and not option.hide_choices:
        if len(_choices) <= 3:
            possible_choices = ", ".join(fmt_choice(_choice) for _choice in _choices)
            choices_help = f"{_markdown_open}(choices are: {possible_choices}){_markdown_close}"
        else:
            sep = " \n - "
            possible_choices = sep + sep.join(fmt_choice(_choice) for _choice in _choices)
            choices_help = f"\n{_markdown_open}Possible choices are:{_markdown_close}" + possible_choices
        return option.help + f" {choices_help}" if option.help else choices_help
    else:
        return option.help



from .base import Formatter
from .utils import FormatterType, get_formatter

try:
    from .rich_formatter import RichFormatter

    HAS_RICH = True
except ImportError:
    RichFormatter = None  # type: ignore[misc, assignment]
    HAS_RICH = False


def get_default_formatter() -> Formatter:
    """Return the best available formatter (RichFormatter if available, else Formatter)."""
    return get_formatter()



from __future__ import annotations
import textwrap
from dataclasses import dataclass, field
from functools import wraps
from inspect import getdoc
from typing import Any, NamedTuple, Callable

from .exceptions import DuplicatedCommandError, CommandException, CommandNotFoundError
from .utils import (
    CommandOption,
    CommandDerivedOption,
    extract_function_info,
    parse_input_args,
    run_function,
    convert_args_to_dict,
)


class ParentArg(NamedTuple):
    cmd: str
    options: list[CommandOption]
    input_options: list[str]
    options_processor: Callable | None = None
    propagate_args: bool = False


ParentArgs = list[ParentArg]


def clean_multiline(s: str) -> str:
    return textwrap.dedent(s).strip()


@dataclass
class Command:
    name: str
    fn: Callable
    help: str | None = None
    options: list[CommandOption] = field(default_factory=list)
    description: str | None = None
    derived_options: list[CommandDerivedOption] = field(default_factory=list)

    @property
    def positional_args(self) -> list[CommandOption]:
        return [opt for opt in self.options if opt.is_positional_arg]

    @property
    def keyword_args(self) -> list[CommandOption]:
        return sorted(
            [opt for opt in self.options if not opt.is_positional_arg],
            key=lambda x: (-x.is_required, x.name),
        )

    @property
    def options_sorted(self) -> list[CommandOption]:
        """Sorts with the following order:
        - positional
        - keyword required
        - keyword optional
        """
        return self.positional_args + self.keyword_args

    def run(self, *args, **kwargs):
        run_function(self.fn, *args, **kwargs)

    def __post_init__(self):
        self.description = clean_multiline(self.description) if self.description else None

        keyword_params = [x for x in self.options if not x.is_positional_arg]
        if not keyword_params:
            return

        _keyword_args = set()
        for _param in keyword_params:
            for _keyword_arg in _param.keyword_args:
                if _keyword_arg in _keyword_args:
                    raise ValueError(f'Duplicate keyword args found "{_keyword_arg}"')
                _keyword_args.add(_keyword_arg)


@dataclass
class CommandMeta:
    cmd_name: str
    fn_args: dict
    cmd_args: dict


OnCommandRun = Callable[[CommandMeta], None]


@dataclass
class CommandGroup:
    """A group of commands that can be used to organize commands and options"""

    name: str | None = None
    """ Name of the command group, used to identify it in the CLI"""
    help: str | None = None
    """ Short line to explain the command group"""
    description: str | None = None
    """ Description of the command group"""
    options_processor: Callable | None = None
    """ Function to process the options before running the command."""
    propagate_options: bool = False
    """ If set to True, the options will be propagated to the sub-commands."""
    on_cmd_run: OnCommandRun | None = None
    """ Function to call when a command is run, with the command name and arguments."""
    # _formatter: Formatter = field(init=False, default=None)
    _options: list[CommandOption] = field(init=False, default_factory=list)
    """ List of options for the command group, can be used to add global options."""
    _commands: dict[str, Command] = field(init=False, default_factory=dict)
    """ Dictionary of commands for the command group, can be used to add commands."""
    _command_groups: dict[str, CommandGroup] = field(init=False, default_factory=dict)
    """ Dictionary of sub-command groups for the command group, can be used to add sub-groups."""
    #

    def __post_init__(self):
        self.description = clean_multiline(self.description) if self.description else None

    @property
    def commands(self) -> dict:
        """Returns a dictionary of commands and command groups, sorted by command name."""
        return {k: self._commands.get(k) or self._command_groups[k] for k in sorted(self.command_names)}

    @property
    def options(self):
        """Returns a list of options, sorted by whether they are required or not."""
        return sorted(self._options, key=lambda x: x.is_required, reverse=True)

    @property
    def options_sorted(self):
        """Alias for options, for consistency with Command interface."""
        return self.options

    def add_sub_parser(self, help: str | None = None, description: str | None = None):
        """Adds a sub-parser to the command group, which can be used to add sub-commands."""
        cls = type(self)
        return cls(help=help, description=description)  # noqa

    def add_option(
        self,
        *args: str,
        help: str | None = None,
        data_type: Any = bool,
        default: Any = False,
    ):
        """Adds an option to the command group."""
        opt = CommandOption(
            default=default,
            help=help,
            keyword_args=args,
        )
        opt.data_type = data_type
        self._options.append(opt)

    @property
    def command_names(self) -> set[str]:
        """Returns a set of command names, including both commands and command groups."""
        return set(self._commands.keys()) | (self._command_groups.keys())

    def add_group(self, group: CommandGroup):
        """Adds a sub-command group to the command group."""
        if group.name is None:
            raise NotImplementedError("A group must have a name")

        if group.name in self.command_names:
            raise DuplicatedCommandError(f"Duplicated command found for {group.name!r}", group.name)
        group.on_cmd_run = self.on_cmd_run
        self._command_groups[group.name] = group

    def add_command(
        self,
        f,
        cmd: str | None = None,
        help: str | None = None,
        description: str | None = None,
        is_main: bool = False,
    ):
        """Adds a command to the command group."""

        cmd_name = "__main__" if is_main else cmd or f.__name__
        if cmd_name in self.commands:
            raise DuplicatedCommandError(f"Duplicated command found for {cmd_name!r}", cmd_name)

        if cmd_name == "__main__" and self.commands:
            raise CommandException("Main command cannot be added with other commands")
        if cmd_name != "__main__" and "__main__" in self.commands:
            raise CommandException(f"Command {cmd_name!r} cannot be added with main command")

        _options, _derived_options = extract_function_info(f)
        self._commands[cmd_name] = Command(
            name=cmd_name,
            fn=f,
            options=_options,
            derived_options=_derived_options,
            help=help,
            description=description or getdoc(f),
        )

    def command(
        self,
        cmd: str | None = None,
        help: str | None = None,
        description: str | None = None,
        is_main: bool = False,
    ):
        """Decorator to mark a function as a command."""

        def _command(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            self.add_command(f, cmd=cmd, help=help, description=description, is_main=is_main)
            return wrapper

        return _command

    def processor(self):
        """Decorator to mark a function as an option processor."""

        def _processor(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            options, _ = extract_function_info(f)
            for option in options:
                self._options.append(option)
            self.set_options_processor(f)
            return wrapper

        return _processor

    def run_with_args(self, *args, parent_args: ParentArgs | None = None):
        """Runs the command with the given arguments."""

        # Collect all global option names from current group and parent args
        global_option_names = set()
        for option in self.options:
            global_option_names.update(option.keyword_args)

        # Add parent global options
        if parent_args:
            for parent_arg in parent_args:
                for option in parent_arg.options:
                    global_option_names.update(option.keyword_args)
        # Splits the input arguments into:
        # - cmd: The command name to execute
        # - global_options: Options that apply to the current command group
        # - cmd_options: Options specific to the sub-command
        cmd, global_options, cmd_options = parse_input_args(args, self.command_names, global_option_names)
        # Determines if cmd refers to a sub-command group or a direct command
        command_group = self._command_groups.get(cmd) if cmd else None
        command = (command_group or self)._commands.get(cmd) if cmd else None

        parent_args = parent_args or []
        # If we have a command group, we need to run it with the given command
        if command_group:
            # Maintains a chain of parent command contexts for nested command structures
            # If it's a command group, recursively calls run_with_args on the sub-group
            if cmd is None:
                raise NotImplementedError('"cmd" cannot be empty')
            parent_args.append(
                ParentArg(
                    cmd,
                    self.options,
                    global_options,
                    options_processor=self.options_processor,
                    propagate_args=self.propagate_options,
                )
            )
            return command_group.run_with_args(*cmd_options, parent_args=parent_args)

        # Checks if help was requested and raises a special exception to display help
        if set(global_options + cmd_options) & {"-h", "--help"}:
            raise ShowHelpError(group=command_group or self, parent_args=parent_args, command=command)

        if not command:
            # When cmd is None (no valid command found), try to find the first
            # non-option argument that the user might have intended as a command
            input_cmd = cmd
            if input_cmd is None:
                for arg in args:
                    if not arg.startswith("-"):
                        input_cmd = arg
                        break
            raise CommandNotFoundError(list(self.command_names), input_command=input_cmd)

        args_dict = {}
        # Creates parallel lists that include:
        # - Command options + current group options + all parent options (if in sub-command)
        # - Command input + global input + all parent inputs (if in sub-command)
        # - Processors from each level
        # - Propagation settings for each level
        options, input_options, processors, propagate_args = (
            # Command options / current group options / parent options (optional)
            [command.options, self._options],
            [cmd_options, global_options],
            [None, self.options_processor],
            [True, self.propagate_options],
        )

        for parent_arg in parent_args:
            options.append(parent_arg.options)
            input_options.append(parent_arg.input_options)
            processors.append(parent_arg.options_processor)
            propagate_args.append(parent_arg.propagate_args)

        # For each level (command → group → parents):
        # - Converts raw input arguments to a dictionary using option definitions
        # - Runs any processor function with the arguments
        # - If propagation is enabled, merges arguments into the final args dictionary
        for _opts, _input_opts, _processor, _propagate in zip(options, input_options, processors, propagate_args):
            try:
                _arg_dict = convert_args_to_dict(_input_opts, _opts)
            except CommandException as e:
                e.cmd = cmd
                raise e
            if _processor:
                _processor(**_arg_dict)
            if _propagate is not False:
                args_dict.update(_arg_dict)

        cmd_args = args_dict.copy()
        for _derived in command.derived_options:
            args_dict = _derived.update_args(args_dict)

        # Optionally calls a monitoring/logging hook
        if self.on_cmd_run:
            full_command_name = ".".join([x.cmd for x in parent_args] + [command.name])
            self.on_cmd_run(CommandMeta(full_command_name, fn_args=args_dict, cmd_args=cmd_args))

        return command.run(**args_dict)

    def set_options_processor(self, fn: Callable):
        """Sets the options processor function for the command group."""
        self.options_processor = fn


class ShowHelpError(Exception):
    """Exception raised to show help for a command or command group."""

    def __init__(
        self,
        group: CommandGroup,
        command: Command | None = None,
        parent_args: ParentArgs | None = None,
    ):
        self.command = command
        self.group = group
        self.parent_args = parent_args
